import hashlib
import inspect
import pickle
from functools import reduce, wraps
from pathlib import Path
from typing import Callable, Optional, Tuple, TypeVar, Union

from platformdirs import user_cache_dir
from typing_extensions import ParamSpec

from .initialize import Target
from .logging import configure_logging

LOGGER = configure_logging(__file__)

P = ParamSpec("P")
R = TypeVar("R")


def memoize(
    cachedir: Optional[Union[Path, str]] = None,
    appname: Optional[Union[str, Tuple[str, ...]]] = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Memoize a function call to disk.

    Args:
        cachedir (Optional[Union[Path, str]], optional): Directory to store
            cached results. If not provided, we use the platform-specific
            user cache directory we get from platformdirs.
        appname (Optional[Union[str, Tuple[str, ...]]], optional): Name of the
            application to use for the cache directory. If not provided and
            cachedir is not provided, an error is raised. It can either be a
            string or a tuple of strings. If a tuple, a subdirectory is created
            for each string in the tuple.

    Returns:
        Callable[[Callable[P, R]], Callable[P, R]]: A decorator that can be
            applied to a function to memoize it.
    """

    if cachedir is None:
        if appname is None:
            raise ValueError("app_name must be specified if cache_dir is not")

        if isinstance(appname, str):
            appname = (appname,)

        cachedir = reduce(lambda x, y: x / y, appname, Path(user_cache_dir()))

    full_cache_dir = Path(cachedir)
    full_cache_dir.mkdir(parents=True, exist_ok=True)

    def _memoize(func: Callable[P, R]) -> Callable[P, R]:

        # get the fully specified function name
        function_name = Target.to_string(func)

        # get a signature for the function; we will bound it to arguments
        # later to derive a hash.
        function_signature = inspect.signature(func)

        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:

            # we accumulate all arguments in a hash here; we also use the
            # fully specified function name to derive a filename at which
            # to cache.
            input_hash = hashlib.sha1()
            input_hash.update(function_name.encode("utf-8"))

            # bound the signature to the arguments; we also build a nice
            # string representation of the arguments for logging purposes.
            bounded_arguments = function_signature.bind(*args, **kwargs)
            arguments_representation = ""

            # we iterate over the arguments and add them to the hash unless
            # the are either a class instance or a function.
            for i, (k, v) in enumerate(bounded_arguments.arguments.items()):

                if i == 0 and (k == "self" or k == "cls"):
                    # we skip cls or self if they are the first argument
                    # provided.
                    continue

                input_hash.update(pickle.dumps((k, v)))
                arguments_representation += f"{k}={v}, "

            # we remove the last comma and space from the string representation
            # we added to the last parameter.
            arguments_representation = arguments_representation.rstrip(", ")
            h = input_hash.hexdigest()

            # this is where we will store/load the cached result to/from.
            cache_file = full_cache_dir / f"{h}.pkl"

            if cache_file.exists():
                # cache hit!
                LOGGER.debug(
                    f"Loading {function_name}({arguments_representation}) "
                    f"from {full_cache_dir} with hash {h}."
                )
                with open(cache_file, "rb") as f:
                    return pickle.load(f)
            else:
                # cache miss!
                result = func(*args, **kwargs)
                LOGGER.debug(
                    f"Loading {function_name}({arguments_representation}) "
                    f"to {full_cache_dir} with hash {h}."
                )
                with open(cache_file, "wb") as f:
                    pickle.dump(result, f)

            return result

        return wrapper

    return _memoize



from typing import Any, Dict, Optional, Sequence, Union

from omegaconf import DictConfig, ListConfig
from rich.console import Console
from rich.table import Column, Table
from rich.traceback import install
from rich.tree import Tree

from .core import traverse
from .utils import SpringsConfig


def add_pretty_traceback():
    # setup nice traceback through rich library
    install(show_locals=SpringsConfig.RICH_LOCALS)


def print_table(
    title: str,
    columns: Sequence[str],
    values: Sequence[Sequence[Any]],
    colors: Optional[Sequence[str]] = None,
):
    colors = list(
        colors or ["magenta", "cyan", "red", "green", "yellow", "blue"]
    )
    if len(columns) > len(colors):
        # repeat colors if we have more columns than colors
        colors = colors * (len(columns) // len(colors) + 1)

    table = Table(
        *(
            Column(column, justify="center", style=color, vertical="middle")
            for column, color in zip(columns, colors)
        ),
        title=f"\n{title}",
        min_width=len(title) + 2,
    )
    for row in values:
        table.add_row(*row)

    Console().print(table)


def print_config_as_tree(title: str, config: Union[DictConfig, ListConfig]):
    def get_parent_path(path: str) -> str:
        return path.rsplit(".", 1)[0] if "." in path else ""

    trees: Dict[str, Tree] = {"": (root := Tree(f"[bold]\n{title}[/bold]"))}
    nodes_order: Dict[str, Dict[str, int]] = {}

    all_nodes = sorted(
        traverse(config, include_nodes=True, include_leaves=False),
        key=lambda spec: spec.path.count("."),
    )
    for spec in all_nodes:
        parent_path = get_parent_path(spec.path)
        tree = trees.get(parent_path, None)
        if spec.key is None or tree is None:
            raise ValueError("Cannot print disjoined tree")

        label = "[bold {color}]{repr}[/bold {color}]".format(
            color="magenta" if isinstance(spec.value, DictConfig) else "cyan",
            repr=spec.key if isinstance(spec.key, str) else f"[{spec.key}]",
        )
        subtree = tree.add(label=label)
        trees[spec.path] = subtree
        nodes_order.setdefault(parent_path, {})[label] = spec.position

    for spec in traverse(config, include_nodes=False, include_leaves=True):
        parent_path = get_parent_path(spec.path)
        tree = trees.get(parent_path, None)
        if tree is None:
            raise ValueError("Cannot print disjoined tree")

        label = (
            f"[bold]{spec.key}[/bold] ({spec.type.__name__}) = {spec.value}"
        )
        tree.add(label=label)
        nodes_order.setdefault(parent_path, {})[label] = spec.position

    for label, tree in trees.items():
        # sort nodes in each tree to match the order the appear in the config
        tree.children.sort(
            key=lambda child: nodes_order[label][str(child.label)]
        )

    Console().print(root)



import importlib.metadata
import inspect
import os
import re
import sys
import warnings
from ast import literal_eval
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, Optional, Type


def clean_multiline(string: str) -> str:
    string = dedent(string.strip())
    string = re.sub(r"\s*\n\s*", " ", string)
    return string


def get_annotations(obj: Any) -> Dict[str, type]:
    # use inspect.get_annotations if python >= 3.10
    if sys.version_info >= (3, 10):
        return inspect.get_annotations(obj)
    else:
        return getattr(obj, "__annotations__", {})


def get_version() -> str:
    """Get the version of the package."""

    # This is a workaround for the fact that if the package is installed
    # in editable mode, the version is not reliability available.
    # Therefore, we check for the existence of a file called EDITABLE,
    # which is not included in the package at distribution time.
    path = Path(__file__).parent / "EDITABLE"
    if path.exists():
        return "dev"

    try:
        # package has been installed, so it has a version number
        # from pyproject.toml
        version = importlib.metadata.version(__package__ or __name__)
    except importlib.metadata.PackageNotFoundError:
        # package hasn't been installed, so set version to "dev"
        version = "dev"

    return version


class SpringsWarnings:
    WARNINGS = literal_eval(os.environ.get("SPRINGS_WARNINGS", "True"))

    @classmethod
    def toggle(cls, value: Optional[bool] = None):
        if value is None:
            value = not cls.WARNINGS
        cls.WARNINGS = value

    @classmethod
    def _warn(
        cls: Type["SpringsWarnings"],
        message: str,
        category: Type[Warning],
        stacklevel: int = 2,
    ):
        if cls.WARNINGS:
            warnings.warn(message, category, stacklevel=stacklevel)

    @classmethod
    def missing_type(cls: Type["SpringsWarnings"], fn_name: str, type_: Any):
        if type_ is None:
            cls._warn(
                clean_multiline(
                    f"""It is strongly recommended to provide a _type_ argument
                        to `{fn_name}`. This ensures that the correct type is
                        annotated as the return value. Further, it performs
                        type checking on the initialized object."""
                ),
                category=UserWarning,
            )

    @classmethod
    def deprecated(
        cls: Type["SpringsWarnings"],
        deprecated: str,
        removed_when: Optional[str] = None,
        replacement: Optional[str] = None,
    ):
        msg = f"`{deprecated}` is deprecated"
        if removed_when:
            msg += f" and will be removed in Springs {removed_when}."
        else:
            msg += " and may be removed in a future version."

        if replacement:
            msg += f" Use `{replacement}` instead."

        cls._warn(message=msg, category=DeprecationWarning)

    @classmethod
    def argument(cls: Type["SpringsWarnings"], arg_name: str, obj_name: str):
        cls._warn(
            f"'{arg_name}' was provided to `{obj_name}`, but it is ignored",
            category=SyntaxWarning,
        )


class SpringsConfig:
    """A class to hold the configuration for Springs"""

    RICH_LOCALS: bool = literal_eval(
        os.environ.get("SPRINGS_RICH_LOCALS", "False")
    )

    def toggle_rich_locals(self, value: Optional[bool] = None):
        if value is None:
            value = not self.RICH_LOCALS
        self.RICH_LOCALS = value
